"""
==================================================
Repo: https://github.com/ksehic/LassoBench
Install by:
git clone https://github.com/ksehic/LassoBench.git
cd LassoBench/
pip install -e .
==================================================
"""

import numpy as np
from HDBOBenchmark.funcs.base.FunctionBase import TestFunction
import LassoBench


class LassoBenchmark(TestFunction):
    def __init__(
        self,
        benchname="synt_simple",
        dim=None,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=False,
    ):
        if "synt_" in benchname:
            self.lassobenchfunc = LassoBench.SyntheticBenchmark(pick_bench=benchname)
        else:
            self.lassobenchfunc = LassoBench.RealBenchmark(pick_data=benchname)

        super().__init__(dim=self.lassobenchfunc.n_features, name=benchname)

        self._x_bound = np.array([[-1, 1]] * self.dim, dtype=float)  # shape(dim, 2)
        y_scale_dict = {
            "Breast_cancer": 6.29,
            "Diabetes": 0.90,
            "DNA": 0.59,
            "leukemia": 7.73,
            "RCV": 0.32,
        }
        min_val_dict = {
            "Breast_cancer": 0.2609,
            "Diabetes": 0.648,
            "DNA": 0.292,
            "leukemia": 0.015,
            "RCV": 0.18,
        }
        self.y_scale = (
            y_scale_dict[benchname] if benchname in y_scale_dict.keys() else 100
        )
        self._min_pt = None
        self._min_val = (
            min_val_dict[benchname] if benchname in min_val_dict.keys() else 0
        )
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        x = np.clip(x, -1, 1)
        if x.ndim == 1:
            return self.lassobenchfunc.evaluate(x)
        else:
            x_shape = x.shape
            x = x.reshape(-1, x.shape[-1])
            result = list()
            for x_slice in x:
                result.append(self.lassobenchfunc.evaluate(x_slice))
            result = np.array(result)
            result = result.reshape(x_shape[:-1])
            return result



import numpy as np
from tqdm import tqdm
from hebo.design_space.design_space import DesignSpace
from hebo.optimizers.hebo import HEBO


n_suggestions = 1


def hebo_optimize(func, bounds, n_iterations, hebo_kwargs=None):
    if hebo_kwargs is None:
        hebo_kwargs = dict()

    ndims = bounds.shape[0]
    params = []
    for i in range(ndims):
        dd = {"type": "num"}
        dd["name"] = "x%d" % i
        dd["lb"] = bounds[i, 0]
        dd["ub"] = bounds[i, 1]
        params.append(dd)
    space = DesignSpace().parse(params)
    opt = HEBO(space, **hebo_kwargs)
    history = np.zeros(n_iterations)
    pbar = tqdm(total=n_iterations)
    for i in range(n_iterations):
        rec = opt.suggest(n_suggestions)
        values = np.zeros((n_suggestions, 1))
        for j in range(n_suggestions):
            values[j, :] = -func(rec.values[j, :])
        opt.observe(rec, values)
        history[i] = -opt.y.min()
        pbar.update()

    best_x, best_y = opt.best_x.values, -opt.best_y
    result = {
        "best_x": best_x,
        "best_y": best_y,
        "x_history": opt.X.values,
        "y_history": -opt.y.squeeze(),
        "best_y_history": history,
    }
    return best_x, best_y, result



import numpy as np
from .FunctionBase import TestFunction


class Trid(TestFunction):
    def __init__(
        self,
        dim,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(dim, "Trid")

        self._x_bound = np.array(
            [[-self.dim * self.dim, self.dim * self.dim]] * self.dim, dtype=float
        )  # shape(dim, 2)
        self.y_scale = 6 * self.dim**3
        self._min_pt = np.zeros(self.dim)
        for i in range(self.dim):
            self._min_pt[i] = (i + 1) * (self.dim - i)
        # self._min_val = self._evaluate(self._min_pt)
        self._min_val = -self.dim * (self.dim + 4) * (self.dim - 1) / 6

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        xi_ = x[..., 0]
        sum = (xi_ - 1) ** 2
        for i in range(1, self.dim):
            xi = x[..., i]
            sum += (xi - 1) ** 2 - xi * xi_
            xi_ = xi
        y = sum
        return y



import numpy as np
from .FunctionBase import TestFunction


class Noise(TestFunction):
    def __init__(
        self,
        dim,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(dim, "Noise")

        self._min_pt = np.zeros(self.dim)
        self._min_val = 0

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        return np.random.normal()



from scipy.optimize import linear_sum_assignment
from HDBOBenchmark.algorithms.add_bo_quasi_nt.utils.getHungarianCostMatrixOrth import (
    get_hungarian_cost_matrix,
)


def orth_to_permutation(a, cost_idx=2, pwr=2):
    """
    Takes in an Orthogonal matrix and returns the 'closest' permutation matrix
    :param a:
    :param cost_idx:
    :param pwr:
    :return row_ind, col_ind:
    """
    # Prelims
    d, p = a.shape
    if d != p:
        raise Exception("Matrix is not a square matrix.")
    # First get the Cost matrix for the assignment
    c = get_hungarian_cost_matrix(a, cost_idx, pwr)
    # min_X \sum_{ij} C_{ij} X_{ij}  s.t.  X_{ij}=1 iff row i is assigned to column j
    row_ind, col_ind = linear_sum_assignment(c)
    return row_ind, col_ind



import torch
from HDBOBenchmark.algorithms.add_bo_quasi_nt.utils.Dist2GP import dist2gp


def aug_kernel(x1, x2, coord, bw, scale):
    """Each of small kernels is only affected by a subset of the coordinates.
    So we need to make sure that the output of the kernel only depends on these quantities. This is what this function
    is doing. Note that we are not adding the noise here.
    Args:
        x1: A 'n1 x ndims' tensor
        x2: A 'n2 x ndims' tensor
        coord:
        bw:
        scale:
    Returns:
        k(x1, x2): (n1, n2)
    """
    if x1.shape[1] == len(coord):
        x1sub = x1
    else:
        x1sub = x1[:, coord]
    if x2.shape[1] == len(coord):
        x2sub = x2
    else:
        x2sub = x2[:, coord]

    d = dist2gp(x1sub, x2sub)  # (n1, n2)
    return scale * torch.exp(-0.5 * d / bw**2)



import numpy as np
import matplotlib.pyplot as plt
import os


def get_linestyle_list(num: int) -> list:
    if num == 1:
        return [("black", "-")]
    if num <= 6:
        return [
            ("b", "-"),
            ("r", "-"),
            ("g", "-"),
            ("m", "--"),
            ("c", "--"),
            ("purple", "--"),
            ("black", "-"),
        ]
    else:
        return [
            ("b", "-"),
            ("r", "-"),
            ("g", "-"),
            ("m", "-"),
            ("c", "-"),
            ("purple", "-"),
            ("b", "--"),
            ("r", "--"),
            ("g", "--"),
            ("m", "--"),
            ("c", "--"),
            ("purple", "--"),
            ("black", "-"),
            ("black", "--"),
        ]


def linear_interpolation(x, x1, y1, x2, y2):
    return y1 + (x - x1) * (y2 - y1) / (x2 - x1)


def cal_std_data(*data):
    """
    data: {'x':array, 'y':array}
    """
    n_data = len(data)
    assert n_data > 0  # no results?

    x = list()
    mean = list()
    std = list()
    std_up = list()
    std_down = list()
    len_data = [len(data[i]["x"]) for i in range(n_data)]
    p = [0 for _ in range(n_data)]
    flag_end = [False for _ in range(n_data)]
    while True:
        # get next x
        next_x = list()
        for i in range(n_data):
            if not flag_end[i]:
                next_x.append(data[i]["x"][p[i]])
            else:
                continue
        if len(next_x) == 0:
            break

        next_x = np.min(next_x)
        x.append(next_x)
        # get next y
        next_y = list()
        for i in range(n_data):
            if (data[i]["x"][p[i]] > next_x) and (p[i] > 0):
                data_y = linear_interpolation(
                    next_x,
                    data[i]["x"][p[i] - 1],
                    data[i]["y"][p[i] - 1],
                    data[i]["x"][p[i]],
                    data[i]["y"][p[i]],
                )
            elif (data[i]["x"][p[i]] < next_x) and (p[i] < len_data[i] - 1):
                data_y = linear_interpolation(
                    next_x,
                    data[i]["x"][p[i]],
                    data[i]["y"][p[i]],
                    data[i]["x"][p[i] + 1],
                    data[i]["y"][p[i] + 1],
                )
            else:
                data_y = data[i]["y"][p[i]]
            next_y.append(data_y)
        next_y = np.array(next_y)
        # cal mean, std
        mean.append(next_y.mean())
        std.append(next_y.std())
        std_up.append(
            np.sqrt(np.square(next_y[np.where(next_y >= mean[-1])] - mean[-1]).mean())
        )
        std_down.append(
            np.sqrt(np.square(next_y[np.where(next_y <= mean[-1])] - mean[-1]).mean())
        )
        # next p[i]
        for i in range(n_data):
            if data[i]["x"][p[i]] == next_x:
                if p[i] == len_data[i] - 1:
                    flag_end[i] = True
                else:
                    p[i] += 1
    return {
        "x": np.array(x),
        "y": np.array(mean),
        "std": np.array(std),
        "std_up": np.array(std_up),
        "std_down": np.array(std_down),
        "name": data[0]["name"],
    }


def plt_init(style="bmh", **kwargs):
    # print(plt.style.available)
    plt.style.use(style)
    # plt.rc('font', family='STFangsong')


def sub_fig(fig, xlabel="x", ylabel="y", **kwargs):
    ax = fig.add_subplot(1, 1, 1)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    return ax


def plot(data, ax, color="b", linestyle="-", **kwargs):
    ax.plot(data["x"], data["y"], linestyle, label=data["name"], color=color, alpha=0.6)
    if "std_up" in data.keys():
        plt.gca().fill_between(
            data["x"],
            data["y"] - data["std_down"],
            data["y"] + data["std_up"],
            facecolor=color,
            alpha=0.2,
        )
    elif "std" in data.keys():
        plt.gca().fill_between(
            data["x"],
            data["y"] - data["std"],
            data["y"] + data["std"],
            facecolor=color,
            alpha=0.2,
        )


def test():
    cwd = "./figs"

    x = np.array(range(100))
    y1 = np.random.randn(100)
    y2 = np.random.randn(100)

    a = cal_std_data(
        {"x": x, "y": y1, "name": "data1"}, {"x": x, "y": y2, "name": "data2"}
    )

    plt_init(style="bmh")
    Color_list = ("b", "r", "g", "m")
    fig = plt.figure()
    ax = sub_fig(fig, xlabel="x", ylabel="y")
    plot(a, ax, color=Color_list[0])
    ax.legend(loc="best", frameon=True, facecolor="white")
    plt.tight_layout()
    try:
        plt.show()
    except:
        os.makedirs(cwd, exist_ok=True)
        save_img_path = os.path.join(cwd, "test.png")
        if os.path.exists(save_img_path):
            os.remove(save_img_path)
        fig.savefig(save_img_path, bbox_inches="tight", dpi=300)


if __name__ == "__main__":
    test()



from HDBOBenchmark.args.base.template import Args as ArgsTemplate
from HDBOBenchmark.funcs import TestFuncs as func_list


class Args(ArgsTemplate):
    def __init__(self):
        super().__init__()
        self.name = "HDBOBenchmark"
        # BO
        self.func = func_list


args = Args()



from typing import Optional
import numpy as np
from .base.FunctionBase import TestFunction


def flatten(lists):
    result = list()
    for o in lists:
        if isinstance(o, list):
            result.extend(flatten(o))
        else:
            result.append(o)
    return result


class AdditiveFunction(TestFunction):
    def __init__(
        self,
        sub_funcs: list,
        name="AdditiveFunction",
        weights=None,
        noisefunc_label_list: Optional[list] = None,
        if_permute=False,
        if_auto_scale=True,
    ):
        """
        Args:
            sub_funcs: list: [func1, func2, ...], Nested is allowed
            weights: list: [w1, w2, ...], Nested is allowed
            if_permute: if permute the order of X
        """
        sub_funcs = flatten(sub_funcs)
        assert len(sub_funcs) > 0
        self.sub_funcs = sub_funcs

        self._dim = list()
        self._x_bound = list()
        for f in sub_funcs:
            self._dim.append(f.dim)
            self._x_bound.append(f.bound)

        self._dim = np.sum(self._dim)
        self._x_bound = np.concatenate(self._x_bound)

        super().__init__(self.dim, name)

        if weights is None:
            self.weights = np.ones(len(sub_funcs))
        else:
            self.weights = flatten(weights)
        assert len(self.weights) == len(sub_funcs)

        if noisefunc_label_list is None:
            self.noisefunc_label_list = [False for _ in range(len(sub_funcs))]
        else:
            self.noisefunc_label_list = flatten(noisefunc_label_list)
        assert len(self.noisefunc_label_list) == len(sub_funcs)

        # if_permute
        self.if_permute = if_permute
        if if_permute:
            self.permute_order = np.random.permutation(self._dim)
            self.invert_order = np.zeros(self._dim).astype(int)
            self.invert_order[self.permute_order] = np.arange(self._dim)

        # extremums
        self._min_pt = list()
        self._min_val = 0
        for f in sub_funcs:
            f_min_pt = f.min_pt
            if f_min_pt is None:
                self._min_pt = None
            if self._min_pt is not None:
                self._min_pt.append(f_min_pt)
            f_min_val = f.min_val
            if f_min_val is None:
                self._min_val = None
            if self._min_val is not None:
                self._min_val += f_min_val

        if self._min_pt is not None:
            self._min_pt = np.concatenate(self._min_pt)
            self._min_pt = self._min_pt[self.invert_order]

        # if_auto_scale
        self.init_scaler()
        self.if_auto_scale = if_auto_scale
        if if_auto_scale:
            weights = 0
            for w, l in zip(self.weights, self.noisefunc_label_list):
                if not l:
                    weights += w
            self.y_scaler[1] /= weights

    def _evaluate(self, x) -> dict:
        result = dict()
        fake_val = 0
        val = 0
        p = 0
        for f, w, l in zip(self.sub_funcs, self.weights, self.noisefunc_label_list):
            dim = f.dim
            y = f(x[p : p + dim]) * w
            val += y
            if l:
                fake_val += y
            p += dim
        result["y"], result["true_y"] = val, val - fake_val
        return result

    def evaluate(self, x) -> dict:
        x = self._x_preprocess(x)
        result = dict()
        result["x"] = x
        if self.if_permute:
            x = x[self.permute_order]
        x = self._cal_descaled(x, self.x_scaler)
        _result = self._evaluate(x)
        y, true_y = self._cal_scaled(_result["y"], self.y_scaler), self._cal_scaled(
            _result["true_y"], self.y_scaler
        )
        if not self.if_minium:
            y, true_y = -y, -true_y
        result["y"], result["true_y"] = y, true_y
        return result

    def __call__(self, x) -> float:
        x = self._x_preprocess(x)
        if self.if_permute:
            x = x[:, self.permute_order]
        x = self._cal_descaled(x, self.x_scaler)
        y = 0
        p = 0
        for f, w in zip(self.sub_funcs, self.weights):
            dim = f.dim
            y += f(x[:, p : p + dim]) * w
            p += dim
        y = self._cal_scaled(y, self.y_scaler)
        if not self.if_minium:
            y = -y
        return y

    def to_torch(self, device="cpu"):
        TestFunction.to_torch(self, device)
        for func in self.sub_funcs:
            func.to_torch(device)



from tensorboard.backend.event_processing import event_accumulator
import json
import os
import re
import pprint


def load_tensorboard_file(tensorboard_path):
    ea = event_accumulator.EventAccumulator(tensorboard_path)
    ea.Reload()
    return ea


def read_tensorboard_data(ea, val_name):
    """读取tensorboard数据，
    tensorboard_path是tensorboard数据地址val_name是需要读取的变量名称"""
    data_dict = {
        "step": list(),
        "value": list(),
        "wall_time": list(),
    }
    data_list = ea.scalars.Items(val_name)
    for scalar in data_list:
        data_dict["step"].append(scalar.step)
        data_dict["value"].append(scalar.value)
        data_dict["wall_time"].append(scalar.wall_time)
    return data_dict


def path_find_file(path, pattern):
    files_name = os.listdir(path)
    for name in files_name:
        file = os.path.join(path, name)
        if os.path.isfile(file) and (re.match(pattern, name) is not None):
            return name
    return None


def log_scalars(logger, result: dict, global_step):
    if logger is not None:
        for n, v in result.items():
            logger.add_scalar(n, v, global_step=global_step)


def log_hparam(args, dir, filename="hparam.json"):
    hparam_dict = _get_hparam_dict(args)
    pprint.pprint(hparam_dict)
    hparam_json = json.dumps(
        hparam_dict, sort_keys=True, indent=4, separators=(",", ": ")
    )
    with open(os.path.join(dir, filename), "w") as f:
        f.write(hparam_json)
    del hparam_dict, hparam_json


def _get_hparam_dict(args):
    def get_hparam_dict(hparam_dict, item, name):
        if isinstance(item, (bool, str, float, int)):
            hparam_dict[name] = item
        elif isinstance(item, (dict,)):
            for key in item:
                get_hparam_dict(hparam_dict, item[key], f"{name}.{key}")
        elif hasattr(item, "__name__"):
            hparam_dict[name] = item.__name__
        elif item is None:
            hparam_dict[name] = "None"
        else:
            hparam_dict[name] = item.__class__.__name__

    hparam_dict = {}
    for name in args.__dict__:
        get_hparam_dict(hparam_dict, getattr(args, name), name)
    return hparam_dict



from typing import Callable
from torch import Tensor
from torch.linalg import solve
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.augKernel import aug_kernel


def compute_ind_gp_output(
    x_test: Tensor,
    x_train: Tensor,
    coord: Tensor,
    ell: Tensor,
    alpha: Tensor,
    bw: Tensor,
    scale: Tensor,
    mean_func: Callable,
):
    """
    This function returns the predictive mean and GP for an individual GP
    Args:
        x_test: A 'n2 x ndims' tensor, testing set.
        x_train: A 'n1 x ndims' tensor, training set.
        coord: coordinates of variables
        ell: A 'n1 x n1' tensor which is cholesky decompostion of kernel matrix of training set.
        alpha:
        bw:  A tensor which gives the bandwidth for current group.
        scale: A tensor which gives the scale for current group.
        mean_func:
    Returns:
        y_mu, y_std, y_k:
    """
    # compute K12 and K22
    k12 = aug_kernel(x_train, x_test, coord, bw, scale)  # (n1, n2)
    k22 = aug_kernel(x_test, x_test, coord, bw, scale)  # (n2, n2)
    """
    Note that we are not adding noise to K22 here
    If we have already observed at x_te then we need to account for each individual noise.
    But if we are interested in prediction at a new unobserved point then we need not add noise.
    """
    # compute the outputs
    y_mu = mean_func(x_test).reshape(-1, 1) + k12.T @ alpha  # Predictive Mean

    v = solve(ell, k12)  # (n1, n2)
    y_k = k22 - v.T @ v  # Predictive Variance  shape(n2, n2)
    return y_mu.squeeze(), y_k.diag()



import torch
from torch import Tensor
from typing import Callable, List
from torch.linalg import solve
from torch.linalg import cholesky
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.combinedKernelNoise import (
    combined_kernel_noise,
)


def combined_rot_mean_func(
    x: Tensor,
    z: Tensor,
    common_mean_func: Callable,
    mean_funcs: List[Callable],
    decomposition: List[Tensor],
):
    """
    The common Mean Func takes in X as its arguments while meanFuncs take in Z = X @ A as its arguments.
    Args:
        x:
        z:
        common_mean_func:
        mean_funcs:
        decomposition:
    Returns:
    """
    mu0 = common_mean_func(x)
    num_groups = len(decomposition)
    for k in range(num_groups):
        coord = decomposition[k]
        mu0 += mean_funcs[k](z[:, coord])
    return mu0


def norm_rot_marg_likelihood(
    sigma_sms: Tensor,
    sigma_prs: Tensor,
    decomposition: List[Tensor],
    a: Tensor,
    train_x: Tensor,
    train_y: Tensor,
    mean_funcs: List[Callable],
    common_mean_func: Callable,
    noises: Tensor,
    common_noise: float,
):
    """
    Returns the normalized marginal likelihood
    Args:
        sigma_sms: A 'numGroups'-dim ndarray, which gives the bandwidth for each kernel.
        sigma_prs: A 'numGroups'-dim ndarray, which gives the scale for each kernel.
        decomposition:
        a: A 'ndims x ndims' ndarray, which is norm rotation transform.
        x: A 'num x ndims' ndarray, which is training set.
        y: A 'num' ndarray, which is target set.
        mean_funcs:
        common_mean_func:
        noises:
        common_noise:
    Returns:
        marg_likelihood:
    """
    num_pts = train_x.shape[0]
    # Apply transformation and compute normalized marginal likelihood
    z = train_x @ a
    k_y = combined_kernel_noise(
        z, decomposition, sigma_sms, sigma_prs, noises, common_noise
    )
    ell = cholesky(k_y)  # Since there is noise, the matrix is nonsingular
    train_y = train_y.unsqueeze(-1)  # (num, 1)
    mean_z = combined_rot_mean_func(
        train_x, z, common_mean_func, mean_funcs, decomposition
    ).reshape(
        -1, 1
    )  # (num, 1)
    y_ = train_y - mean_z
    alpha = solve(ell.T, solve(ell, y_))  # inv(ell ell.T) @ (y-m_x)
    return (
        -0.5 * torch.squeeze(y_.T @ alpha)
        - torch.sum(torch.log(torch.diag(ell)))
        - num_pts / 2 * torch.log(torch.tensor(2 * torch.pi))
    )



import torch
from torch import Tensor
from typing import Callable
from scipy.stats import norm


def get_ucb_utility(x: Tensor, func_h: Callable, num_evals: int):
    # Prelims
    ndims = x.shape[1]  # Expecting each x to be a row vector here.
    # Set beta_t. Using recommendation from Section 6 in Srinivas et al., ICML 2010
    t = torch.tensor(num_evals + 1)
    # Linear in dims, log in t
    beta_t = ndims * torch.log(2 * t) / 5
    # Obtain mean and standard deviation
    mu, var = func_h(x)  # shape(batch, 1)
    if (var < 0).any():
        sqrt_var = torch.where(
            var < 0, -torch.sqrt(torch.abs(var)), torch.sqrt(torch.abs(var))
        )
    else:
        sqrt_var = torch.sqrt(var)
    return torch.squeeze(mu + torch.sqrt(beta_t) * sqrt_var)  # shape(batch)


def truncate_gaussian_mean(mu, sigma, trunc):
    """
    Computes the value E[max(0,x)] where x~N(mu, sigma**2)
    :param mu:
    :param sigma:
    :param trunc:
    :return trunc_mean:
    """
    y = mu - trunc
    var_zero_idxs = sigma == 0
    return var_zero_idxs * max(y, 0) + (~var_zero_idxs) * (
        y * norm.cdf(y / sigma) + sigma * norm.pdf(y / sigma)
    )


def get_ei_utility(x, gp_func_h, trunc):
    """
    Expected Improvement Utility. Applies only to non-additive functions.
    :param x:
    :param gp_func_h:
    :param trunc:
    :return:
    """
    mu, sigma = gp_func_h(x)
    return truncate_gaussian_mean(mu, sigma, trunc)



import logging

import numpy as np
from numpy.linalg import cholesky


def stable_cholesky(k):
    """
    Sometimes nominally positive semi-definite matrices are not  positive semi-definite matrices due to numerical
    issues. By adding a small value to the diagonal we can make it positive semi-definite. This is what this
    function does.
    Use this iff you know that k should be positive semi-definite. We do not check for errors.
    :param k: matrix
    :return l: lower-triangular Cholesky factor of k
    """
    diag_power = min(np.ceil(np.log10(np.abs(np.min(np.diag(k))))) - 1, -11)
    if not (np.abs(diag_power) < np.inf):
        diag_power = -10

    # Now keep trying until Cholesky is successful
    success = False
    k += 10**diag_power * np.eye(k.shape[0])
    while not success:
        try:
            l = cholesky(k)
            success = True
        except Exception:
            logging.warning("CHOL failed with diag_power = %d\n" % diag_power)
            diag_power += 1
            k += 10**diag_power * np.eye(k.shape[0])
    return l



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys
import numpy as np
from .param import Parameter


class IntExponentPara(Parameter):
    def __init__(self, param_dict):
        """
        Integer value, search in log-scale, and the exponent must be integer.
        For example, parameter whose values must be one of [32, 64, 128, 512, 1024]
        """
        super().__init__(param_dict)
        self.base = param_dict["base"]
        self.lb = np.round(np.log(param_dict["lb"]) / np.log(self.base))
        self.ub = np.round(np.log(param_dict["ub"]) / np.log(self.base))

    def sample(self, num=1):
        if num <= 0:
            raise ValueError(f"num <= 0: {num}")
        exponent = np.random.randint(self.lb, self.ub + 1, num)
        return self.base**exponent

    def transform(self, x):
        return np.log(x) / np.log(self.base)

    def inverse_transform(self, x):
        return (self.base**x).astype(int)

    @property
    def is_numeric(self):
        return True

    @property
    def opt_lb(self):
        return self.lb

    @property
    def opt_ub(self):
        return self.ub

    @property
    def is_discrete(self):
        return True

    @property
    def is_discrete_after_transform(self):
        return True



from HDBOBenchmark.args.base.template import Args as ArgsTemplate
from HDBOBenchmark.funcs.realistic.HPOBench import HPOBench as get_func


class Args(ArgsTemplate):
    def __init__(self):
        super().__init__()
        self.name = "HPOBench"
        # BO

        self.get_func = get_func
        self.func = get_func(
            benchname="ml.XGBoostBenchmark",
            seed=1,
        )  # https://github.com/automl/HPOBench/wiki/Available-Containerized-Benchmarks
        self.dims = self.func.dim
        self.method_kwargs["hebo"]["model_name"] = "gumbel"


args = Args()



import numpy as np
from .FunctionBase import TestFunction


class Shubert(TestFunction):
    def __init__(
        self,
        dim,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(2, "Shubert")

        self._x_bound = np.array([[-10, 10]] * self.dim, dtype=float)  # shape(dim, 2)
        self.y_scale = 17 * self.dim
        self._min_pt = None  # ** can not find the exact min points **
        self._min_val = -16

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        sum = 0
        for j in range(self.dim):
            xj = x[j]
            for i in range(5):
                sum += i * np.cos((i + 1) * xj + i)
        y = sum
        return y



from HDBOBenchmark.algorithms.add_bo_quasi_nt.utils.getRandPermMat import (
    get_rand_perm_mat,
)


def decomp_opt_partial(func, ndims):
    num_trials = 5
    a = get_rand_perm_mat(ndims).T
    curr_best_val = func(a)
    for i in range(num_trials):
        p = get_rand_perm_mat(ndims).T
        val = func(p)
        if val < curr_best_val:
            a = p
            curr_best_val = val
    return a



import numpy as np
from ..FunctionBase import TestFunction


class VizierBBOBWrapper(TestFunction):
    def __init__(
        self,
        func,
        dim,
        seed: int = 0,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=False,
    ):
        super().__init__(dim, func.__name__)

        self.func = func
        self.seed = seed
        self._x_bound = np.array([[-5, 5]] * self.dim)  # shape(dim, 2)
        self.y_scale = 1
        self._min_pt = np.zeros(self.dim)
        # self._min_val = self._evaluate(self._min_pt)
        self._min_val = 0

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(
        self, x: np.ndarray
    ) -> float:  # black-box objective function to minimize
        assert x.ndim <= 2
        if x.ndim == 2:  # batch integration
            y = list()
            for x_ in x:
                y.append(self.func(x_, self.seed))
            return np.array(y)
        return self.func(x, self.seed)

    def to_torch(self, device="cpu"):
        raise NotImplementedError("BBOB is implemented with Numpy.")

    def set_seed(self, seed: int):
        self.seed = seed



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import pandas as pd
import torch
from torch import Tensor

from .numeric_param import NumericPara
from .integer_param import IntegerPara
from .pow_param import PowPara
from .categorical_param import CategoricalPara
from .bool_param import BoolPara
from .pow_integer_param import PowIntegerPara
from .int_exponent_param import IntExponentPara
from .step_int import StepIntPara


class DesignSpace:
    def __init__(self):
        self.para_types = {}
        self.register_para_type("num", NumericPara)
        self.register_para_type("pow", PowPara)
        self.register_para_type("pow_int", PowIntegerPara)
        self.register_para_type("int_exponent", IntExponentPara)
        self.register_para_type("int", IntegerPara)
        self.register_para_type("step_int", StepIntPara)
        self.register_para_type("cat", CategoricalPara)
        self.register_para_type(
            "bool", CategoricalPara
        )  # consider boolean as categorical for combinatorial opt.
        self.paras = {}
        self.para_names = []
        self.numeric_names = []
        self.enum_names = []

    @property
    def num_paras(self):
        return len(self.para_names)

    @property
    def num_numeric(self):
        return len(self.numeric_names)

    @property
    def num_categorical(self):
        return len(self.enum_names)

    def parse(self, rec):
        self.para_config = rec
        self.named_para_config = {}
        self.paras = {}
        self.para_names = []
        for item in rec:
            if item["type"] not in self.para_types:
                raise ValueError(
                    f"item {item['type']} not in para_types {self.para_types}"
                )
            param = self.para_types[item["type"]](item)
            self.paras[param.name] = param
            if param.is_categorical:
                self.enum_names.append(param.name)
            else:
                self.numeric_names.append(param.name)
            self.named_para_config[item["name"]] = item
        self.para_names = self.numeric_names + self.enum_names
        return self

    def register_para_type(self, type_name, para_class):
        """
        User can define their specific parameter type and register the new type
        using this function
        """
        self.para_types[type_name] = para_class

    def sample(self, num_samples=1):
        """
        df_suggest: suggested initial points
        """
        df = pd.DataFrame(columns=self.para_names)
        for c in df.columns:
            df[c] = self.paras[c].sample(num_samples)
        return df

    def transform(self, data: pd.DataFrame) -> (Tensor, Tensor):
        """
        input: pandas dataframe
        output: xc and xe
        transform data to be within [opt_lb, opt_ub]
        """
        xc = data[self.numeric_names].values.astype(float).copy()
        xe = data[self.enum_names].values.copy()
        if xc.ndim == 1:
            for i, name in enumerate(self.numeric_names):
                xc[None, i] = self.paras[name].transform(xc[None, i])
        else:
            for i, name in enumerate(self.numeric_names):
                xc[:, i] = self.paras[name].transform(xc[:, i])
        if xe.ndim == 1:
            for i, name in enumerate(self.enum_names):
                xe[None, i] = self.paras[name].transform(xe[None, i])
        else:
            for i, name in enumerate(self.enum_names):
                xe[:, i] = self.paras[name].transform(xe[:, i])
        return torch.FloatTensor(xc), torch.LongTensor(xe.astype(int))

    def inverse_transform(self, x: Tensor, xe: Tensor) -> pd.DataFrame:
        """
        input: x and xe
        output: pandas dataframe
        """
        with torch.no_grad():
            inv_dict = {}
            for i, name in enumerate(self.numeric_names):
                inv_dict[name] = self.paras[name].inverse_transform(
                    x.detach().double().numpy()[:, i]
                )
            for i, name in enumerate(self.enum_names):
                inv_dict[name] = self.paras[name].inverse_transform(
                    xe.detach().numpy()[:, i]
                )
            return pd.DataFrame(inv_dict)

    @property
    def opt_lb(self):
        lb_numeric = [self.paras[p].opt_lb for p in self.numeric_names]
        lb_enum = [self.paras[p].opt_lb for p in self.enum_names]
        return torch.tensor(lb_numeric + lb_enum)

    @property
    def opt_ub(self):
        ub_numeric = [self.paras[p].opt_ub for p in self.numeric_names]
        ub_enum = [self.paras[p].opt_ub for p in self.enum_names]
        return torch.tensor(ub_numeric + ub_enum)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__



# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.

# This program is free software; you can redistribute it and/or modify it under
# the terms of the MIT license.

# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the MIT License for more details.

import sys
import numpy as np
from .param import Parameter


class PowPara(Parameter):
    def __init__(self, param_dict):
        super().__init__(param_dict)
        self.base = param_dict.get("base", 10.0)
        self.lb = np.log(param_dict["lb"]) / np.log(self.base)
        self.ub = np.log(param_dict["ub"]) / np.log(self.base)

    def sample(self, num=1):
        if num <= 0:
            raise ValueError(f"num <= 0: {num}")
        return self.base ** np.random.uniform(self.lb, self.ub, num)

    def transform(self, x):
        return np.log(x) / np.log(self.base)

    def inverse_transform(self, x):
        return self.base**x

    @property
    def is_numeric(self):
        return True

    @property
    def opt_lb(self):
        return self.lb

    @property
    def opt_ub(self):
        return self.ub

    @property
    def is_discrete(self):
        return False

    @property
    def is_discrete_after_transform(self):
        return False



import numpy as np
import torch
from typing import Optional, Union
from .FunctionBase import TestFunction


class Sphere(TestFunction):
    def __init__(
        self,
        dim,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(dim, "Sphere")

        self._x_bound = np.array([[-5, 5]] * self.dim, dtype=float)  # shape(dim, 2)
        self.y_scale = 25 * self.dim
        self._min_pt = np.zeros(self.dim)
        # self._min_val = self._evaluate(self._min_pt)
        self._min_val = 0

        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(
        self, x: Union[torch.Tensor, np.ndarray]
    ):  # black-box objective function to minimize
        if self.if_torch:
            y = torch.pow(x[..., : self.dim], 2).sum(-1)
        else:
            y = np.power(x[..., : self.dim], 2).sum(-1)
        return y



class Args:
    def __init__(self):
        self.name = None
        # BO
        self.dims = 10
        self.func = None

        # main
        self.numIterations = 100
        self.n_init_sample = 20

        # HEBO
        # surrogate model: svidkl, svgp, gp, gpy, gpy_mlp, rf, deep_ensemble, psgld, mcbn, masked_deep_ensemble,fe_deep_ensemble,gumbel,catboost
        # surrogate model: gp, gpy, rf, fe_deep_ensemble, gumbel

        self.method_kwargs = dict()
        self.method_kwargs["botorch"] = {
            "n_init": self.n_init_sample,
        }

        self.method_kwargs["hebo"] = {
            "model_name": "gpy",
            "rand_sample": self.n_init_sample,
            # "acq_cls": MACE,
            "es": "nsga2",
            "scramble_seed": None,
        }
        if self.method_kwargs["hebo"]["model_name"] == "gp":
            cfg = {
                "lr": 0.01,
                "num_epochs": 100,
                "verbose": False,
                "noise_lb": 8e-4,
                "pred_likeli": False,
            }
        elif self.method_kwargs["hebo"]["model_name"] == "gpy":
            cfg = {"verbose": False, "warp": True}
        elif self.method_kwargs["hebo"]["model_name"] == "gpy_mlp":
            cfg = {"verbose": False}
        elif self.method_kwargs["hebo"]["model_name"] == "rf":
            cfg = {"n_estimators": 20}
        else:
            cfg = dict()
        self.method_kwargs["hebo"]["model_config"] = cfg

        # TurBO
        self.method_kwargs["turbo"] = {
            "n_init": self.n_init_sample,
            "batch_size": 1,
        }

        # RemBO
        self.method_kwargs["rembo"] = {
            "n_init": self.n_init_sample,
            "em_dim": 4,  # this is strict
        }
        # RemBO
        self.method_kwargs["saasbo"] = {
            "n_init": self.n_init_sample,
            "batch_size": 1,
        }
        # AleBO
        self.method_kwargs["alebo"] = {
            "n_init": self.n_init_sample,
            "d": 7,
        }



import torch
import math
from tqdm import tqdm
from botorch.models import SingleTaskGP
from botorch.fit import fit_gpytorch_mll
from gpytorch.mlls import ExactMarginalLogLikelihood
from botorch.acquisition import UpperConfidenceBound, qExpectedImprovement
from botorch.optim import optimize_acqf

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
dtype = torch.double


class RemBo:
    def __init__(self, d_orig, d_embedding, n_trials, initial_random_samples=10):
        self.initial_random_samples = initial_random_samples
        self.d_embedding = d_embedding
        self.A = torch.randn(size=(d_embedding, d_orig), dtype=dtype, device=device)
        # buffer
        self.train_x = torch.empty(size=(n_trials, d_orig), dtype=dtype, device=device)
        self.x_embedded = torch.empty(
            size=(n_trials, d_embedding), dtype=dtype, device=device
        )
        self.train_y = torch.empty(size=(n_trials, 1), dtype=dtype, device=device)
        self.count = 0

    def select_query_point(self, batch_size=1):
        """
        :param batch_size: the number of query points per time
        :return x_query: (batch_size, d_orig) numpy array
        :return candidate: (batch_size, d_e) numpy array
        """
        embedding_bounds = torch.tensor(
            [[-math.sqrt(self.d_embedding), math.sqrt(self.d_embedding)]]
            * self.d_embedding,
            dtype=dtype,
            device=device,
        )
        if self.count < self.initial_random_samples:
            # Select query point randomly from embedding_boundaries
            candidate = torch.rand(
                size=(1, self.d_embedding), dtype=dtype, device=device
            )
            # manifold
            candidate = (
                candidate * (embedding_bounds[:, 1] - embedding_bounds[:, 0])
                + embedding_bounds[:, 0]
            )
        else:
            train_Y = self.train_y[: self.count]
            train_Y = (train_Y - train_Y.mean()) / (train_Y.std() + 1e-6)
            # fit model
            gp = SingleTaskGP(train_X=self.x_embedded[: self.count], train_Y=train_Y)
            mll = ExactMarginalLogLikelihood(gp.likelihood, gp)
            fit_gpytorch_mll(mll)
            # Construct an acquisition function
            ei = qExpectedImprovement(gp, train_Y.max())
            # ucb = UpperConfidenceBound(gp, beta=0.2)
            # Optimize the acquisition function
            embedding_bounds = embedding_bounds.T
            candidate, acq_value = optimize_acqf(
                acq_function=ei,
                bounds=embedding_bounds,
                q=batch_size,
                num_restarts=10,
                raw_samples=512,
            )  # shape(batch_size, d_e)

        # Map to higher dimensional space and clip to hard boundaries [-1, 1]
        x_query = torch.clip(self._manifold_to_dataspace(candidate), min=-1.0, max=1.0)
        return x_query, candidate

    def _manifold_to_dataspace(self, x_embedded):
        """
        Map data from manifold to original data space.

        :param x_embedded: (1 x d_embedding) numpy.array
        :return: (1 x d_orig) numpy.array
        """
        return x_embedded @ self.A

    def save_point(self, x_query, y_query, x_embedded):
        """Update internal model for observed (X, y) from true function.
        The function is meant to be used as follows.
            1. Call 'select_query_point' to update self.X_embedded with a new
                embedded query point, and to return a query point X_query in the
                original (unscaled) search space
            2. Evaluate X_query to get y_query
            3. Call this function ('update') to update the surrogate model (e.g.
                Gaussian Process)

        Args:
            x_query ((1,d_orig) tensor):
                Point in original input space to query
            y_query (float):
                Value of black-box function evaluated at X_query
            x_embedded ((1,d_e) numpy array):
        """
        # add new rows of data
        self.train_x[self.count] = x_query
        self.train_y[self.count] = y_query
        self.x_embedded[self.count] = x_embedded
        self.count += 1


def rembo(eval_objective, dim, em_dim, n_init, total_trials):
    opt = RemBo(dim, em_dim, total_trials, initial_random_samples=n_init)

    pbar = tqdm(total=total_trials)
    for _ in range(total_trials):
        x_queries, x_embedded = opt.select_query_point()

        # Evaluate the batch of query points 1-by-1
        for x_query, x_e in zip(x_queries, x_embedded):
            y_query = eval_objective((x_query + 1) / 2)  # manifold to [0,1]^D
            opt.save_point(x_query, y_query, x_e)
        pbar.update()

    X, Y = (opt.train_x + 1) / 2, opt.train_y
    return X, Y


# if __name__ == "__main__":
#     import numpy as np

#     DIM = 100
#     EM_DIM = 3
#     N_INIT = 5
#     TOTAL_TRIALS = 30
#     from botorch.test_functions import Branin
#     branin = Branin().to(dtype=dtype, device=device)

#     def branin_emb(x):
#         """x is assumed to be in [0, 1]^d"""
#         lb, ub = branin.bounds
#         return branin(lb + (ub - lb) * x[..., :2]) * -1  # Flip the value for minimization

#     X, Y = rembo(branin_emb, D=DIM, d=EM_DIM, n_init=N_INIT, total_trials=TOTAL_TRIALS)
#     Y_np = -1 * Y
#     from matplotlib import pyplot as plt

#     fig = plt.figure(figsize=(12, 6))
#     ax = fig.add_subplot(111)
#     ax.grid(alpha=0.2)
#     ax.plot(range(1, 31), np.minimum.accumulate(Y_np))
#     ax.plot([0, len(Y_np)], [0.398, 0.398], "--", c="g", lw=3, label="Optimal value")
#     ax.set_xlabel('Iteration')
#     ax.set_ylabel('Best objective found')
#     plt.savefig("results.png")
if __name__ == "__main__":
    from botorch.test_functions import Ackley

    true_dim = 10
    fun = Ackley(dim=true_dim).to(dtype=dtype, device=device)
    fun.bounds[0, :].fill_(-5)
    fun.bounds[1, :].fill_(10)
    DIM = 100
    EM_DIM = 4
    N_INIT = 5
    TOTAL_TRIALS = 30

    def eval_objective(x):
        """This is a helper function we use to unnormalize and evalaute a point"""
        x = x[:true_dim]
        lb, ub = fun.bounds
        return fun(lb + (ub - lb) * x) * -1

    X, Y = rembo(
        eval_objective, dim=DIM, em_dim=EM_DIM, n_init=N_INIT, total_trials=TOTAL_TRIALS
    )
    Y_np = -1 * Y
    import numpy as np
    import matplotlib.pyplot as plt

    fx = np.minimum.accumulate(Y_np)
    plt.plot(fx, marker="", lw=3)

    plt.plot([0, len(Y_np)], [fun.optimal_value, fun.optimal_value], "k--", lw=3)
    plt.show()



import numpy as np

# pip install ax-platform
from HDBOBenchmark.algorithms.alebo import alebo


def alebo_optimize(func, bounds, n_iterations, alebo_kwargs=None):
    if alebo_kwargs is None:
        alebo_kwargs = dict()

    lb, ub = bounds[:, 0], bounds[:, 1]
    if not "D" in alebo_kwargs:
        alebo_kwargs["D"] = func.dim

    def eval_objective_wrap(parameterization):
        """x is assumed to be in [0, 1]^d"""
        x = np.array(
            [parameterization.get(f"x{i}") for i in range(len(parameterization))]
        )
        x = lb + (ub - lb) * x
        y = func(x)  # Flip the value for minimization
        return {"objective": (y, 0.0)}

    X, Y = alebo(
        eval_func=eval_objective_wrap,
        total_trials=n_iterations,
        **alebo_kwargs,
    )

    best_x, best_y = X[np.argmax(Y)], np.max(Y)
    result = {"best_x": best_x, "best_y": best_y, "x_history": X, "y_history": Y}
    return best_x, best_y, result



from HDBOBenchmark.args.base.template import Args as ArgsTemplate
from HDBOBenchmark.funcs.realistic.lassobench import LassoBenchmark as get_func


class Args(ArgsTemplate):
    def __init__(self, benchname="synt_simple"):
        super().__init__()
        self.name = "LassoBench"
        # BO
        self.dims = 100
        self.get_func = get_func
        self.func = get_func(
            benchname=benchname
        )  # synt_simple, synt_medium, synt_high, synt_hard, Breast_cancer, Diabetes, Leukemia, DNA, RCV1

        self.method_kwargs["hebo"][
            "model_name"
        ] = "gumbel"  # gp for dim 5, gumbel for dim 10 20


args = Args(benchname="synt_simple")



import torch
from torch import Tensor
from typing import Callable
from .initializers import gen_batch_initial_conditions, gen_mll_initial_conditions
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_generation.gen import (
    gen_candidates_scipy,
    gen_mll_candidates_scipy,
)


def optimize_acqf(
    acq_function: Callable,
    bounds: Tensor,
    q: int,
    num_restarts: int,
    raw_samples: int,
    return_best_only=True,
):
    """
    Args:
        acq_function: An AcquisitionFunction.
        bounds: A '2 x d' ndarray of lower and upper bounds for each column of 'X'.
        q: The number of candidates.
        num_restarts: The number of starting points for multistart acquisition
            function optimization.
        raw_samples: The number of samples for initialization.
        return_best_only: If False, outputs the solutions corresponding to all
            random restart initializations of the optimization.

    Returns:
        A two-element tuple containing

        - a 'num_restarts x q x d'-dim ndarray of generated candidates.
        - a 'num_restarts'-dim ndarray of associated acquisition values.
    """
    batch_initial_conditions = gen_batch_initial_conditions(
        acq_function=acq_function,
        bounds=bounds,
        q=q,
        num_restarts=num_restarts,
        raw_samples=raw_samples,
    )
    batch_candidates, batch_acq_values = gen_candidates_scipy(
        initial_conditions=batch_initial_conditions,
        acquisition_function=acq_function,
        lower_bounds=bounds[0],
        upper_bounds=bounds[1],
    )
    if return_best_only:
        best = torch.argmax(batch_acq_values.view(-1), dim=0)
        batch_candidates = batch_candidates[best]  # shape(q, ndims)
        batch_acq_values = batch_acq_values[best]  # shape()

    return batch_candidates, batch_acq_values


def optimize_mll(
    mll: Callable,
    bounds: Tensor,
    q: int,
    num_restarts: int,
    raw_samples: int,
    return_best_only=True,
):
    batch_initial_conditions = gen_mll_initial_conditions(
        mll=mll, bounds=bounds, q=q, num_restarts=num_restarts, raw_samples=raw_samples
    )
    batch_candidates, batch_mll_values = gen_mll_candidates_scipy(
        initial_conditions=batch_initial_conditions,
        mll=mll,
        lower_bounds=bounds[0],
        upper_bounds=bounds[1],
    )
    if return_best_only:
        best = torch.argmax(batch_mll_values.view(-1), dim=0)
        batch_candidates = batch_candidates[best]  # shape(q, ndims)
        batch_mll_values = batch_mll_values[best]  # shape()

    return batch_candidates, batch_mll_values



"""
==================================================
Repo: https://github.com/zwicker-group/py-pde
Doc: https://py-pde.readthedocs.io/en/latest/index.html
Install by:
pip install py-pde
==================================================
"""

import numpy as np
from HDBOBenchmark.funcs.base.FunctionBase import TestFunction
from pde import PDE, FieldCollection, ScalarField, UnitGrid


class Brusselator(TestFunction):
    """
    https://py-pde.readthedocs.io/en/latest/examples_gallery/pde_brusselator_expression.html#sphx-glr-examples-gallery-pde-brusselator-expression-py
    References: https://github.com/bhouri0412/rpn_bo
    """

    def __init__(
        self,
        dim=4,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=False,
    ):
        super().__init__(4, "Brusselator")
        self._grid = UnitGrid([64, 64])

        self._x_bound = np.array(
            [[0.1, 5], [0.1, 5], [0.01, 5], [0.01, 5]], dtype=float
        )  # shape(dim, 2)
        self.y_scale = 3
        self._min_pt = None
        self._min_val = 0
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)

    def _evaluate(self, x):  # black-box objective function to minimize
        a, b, d0, d1 = x[0], x[1], x[2], x[3]
        eq = PDE(
            {
                "u": f"{d0} * laplace(u) + {a} - ({b} + 1) * u + u**2 * v",
                "v": f"{d1} * laplace(v) + {b} * u - u**2 * v",
            }
        )
        u = ScalarField(self._grid, a, label="Field $u$")
        v = b / a + 0.1 * ScalarField.random_normal(self._grid, label="Field $v$")
        state = FieldCollection([u, v])
        sol = eq.solve(state, t_range=20, dt=1e-3)
        y = sol.data.mean(
            axis=0
        ).max()  # maximum average density of u,v in a place at the last time t
        del eq, u, v, state, sol
        return y



import torch
from torch import Tensor
from typing import List
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.computeIndGPOutput import (
    compute_ind_gp_output,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.addGPLibkky.computeCombGPOutput import (
    compute_comb_gp_output,
)
from HDBOBenchmark.algorithms.add_bo_quasi_nt.my_optim.optimize import optimize_acqf
from .getAcqUtility import get_ucb_utility, get_ei_utility


def get_next_query_pt(
    train_x: Tensor,
    train_y: Tensor,
    params,
    decomposition: List[Tensor],
    bounds: Tensor,
):
    """Obtain the next query point.
    Args:
        train_x: A 'num x ndims'-dim ndarray, training set.
        train_y: A 'num'-dim ndarray, testing set.
        params: Hyper parameters
        decomposition:
        bounds: A 'ndims x 2' tensor. bounds[:, 0] gives the lower bounds for x,
            and bounds[:, 1] gives the upper bounds for x.
    Returns:
        next_point, next_point_mean, next_point_std:
    """
    # So here's the plane here. We will define a utility function over each GP and
    # pick the point that maximizes this utility function.
    num_groups = len(decomposition)
    ndims = bounds.shape[0]
    sigma_sms, sigma_prs, mean_funcs, common_mean_func = (
        params.sigma_sms,
        params.sigma_prs,
        [params.mean_funcs] * num_groups,
        params.common_mean_func,
    )
    ell, alpha = params.ell, params.alpha
    next_point = torch.zeros(ndims, dtype=torch.double)

    for k in range(num_groups):
        coord = decomposition[k]
        curr_bounds = bounds[coord]
        bw, scale, curr_mean_func = sigma_sms[k], sigma_prs[k], mean_funcs[k]

        def ucb(t):
            return get_ucb_utility(
                t,
                lambda x_test: compute_ind_gp_output(
                    x_test, train_x, coord, ell, alpha, bw, scale, curr_mean_func
                ),
                train_y.shape[0],
            )

        def ei(t):
            return get_ei_utility(
                t,
                lambda x_test: compute_ind_gp_output(
                    x_test, train_x, coord, ell, alpha, bw, scale, curr_mean_func
                ),
                torch.max(train_y),
            )

        if params.utility_func == "UCB":
            utility = ucb
        elif params.utility_func == "EI":
            # Only applies to non-additive models. (i.e. d = D, M = 1)
            utility = ei
        else:
            raise NotImplementedError
        # To maximize the acquisition function
        candidate, acq_value = optimize_acqf(
            utility, curr_bounds.T, q=1, num_restarts=256, raw_samples=256
        )
        candidate = candidate.squeeze()  # (1, ndims) => (ndims)
        # Store next_point in relevant coordinates
        next_point[coord] = candidate
    # Finally return the mean and the standard deviation at next_point
    # Now the combined GP output
    next_point_mean, next_point_std, _ = compute_comb_gp_output(
        next_point.unsqueeze(0),
        train_x,
        decomposition,
        ell,
        alpha,
        sigma_sms,
        sigma_prs,
        mean_funcs,
        common_mean_func,
    )
    return next_point, next_point_mean, next_point_std



from .AdditiveFunction import AdditiveFunction
from .base import *

TestFuncs100 = [
    AdditiveFunction(
        name="0. Sphere(2)_Zero(98)",
        if_permute=True,
        sub_funcs=[Sphere(2), Zero(98)],
        noisefunc_label_list=[False, True],
    ),
    AdditiveFunction(
        name="1. RosenBrock(4)^10_Eggholder(2)^10_Zero(40)",
        if_permute=True,
        sub_funcs=[[[Rosenbrock(4), EggHolder(2)] for _ in range(10)], Zero(40)],
        noisefunc_label_list=[[[False, False] for _ in range(10)], True],
    ),
    AdditiveFunction(
        name="2. ProductSines(50)_RosenBrock(50)",
        if_permute=True,
        sub_funcs=[ProductSines(50), Rosenbrock(50)],
    ),
    AdditiveFunction(
        name="3. Eggholder(2)^50",
        if_permute=True,
        sub_funcs=[EggHolder(2) for _ in range(50)],
    ),
    AdditiveFunction(
        name="4. Rastrigin(100)",
        if_permute=True,
        sub_funcs=[
            Rastrigin(100),
        ],
    ),
    AdditiveFunction(
        name="5. Ackley(10)^10",
        if_permute=True,
        sub_funcs=[Ackley(10) for _ in range(10)],
    ),
    AdditiveFunction(
        name="6. Ackley(20)_ProductSines(20)_Hartmann6d(6)^5_Noise(30)",
        if_permute=True,
        sub_funcs=[
            Ackley(20),
            ProductSines(20),
            [Hartmann6d(6) for _ in range(5)],
            Noise(30),
        ],
        noisefunc_label_list=[False, False, [False for _ in range(5)], True],
    ),
    AdditiveFunction(
        name="7. Ackley(20)_ProductSines(20)_0.5Rastrigin(20)_0.2Hartmann6d(6)^6_Noise(4)",
        if_permute=True,
        sub_funcs=[
            Ackley(20),
            ProductSines(20),
            Rastrigin(20),
            [Hartmann6d(6) for _ in range(6)],
            Noise(4),
        ],
        weights=[1, 1, 0.5, [0.2 for _ in range(6)], 1],
        noisefunc_label_list=[False, False, False, [False for _ in range(6)], True],
    ),
    AdditiveFunction(
        name="8. Ackley(20)_ProductSines(20)_Rastrigin(20)_Hartmann6d(6)^6_Noise(4)",
        if_permute=True,
        sub_funcs=[
            Ackley(20),
            ProductSines(20),
            Rastrigin(20),
            [Hartmann6d(6) for _ in range(6)],
            Noise(4),
        ],
        noisefunc_label_list=[False, False, False, [False for _ in range(6)], True],
    ),
    AdditiveFunction(
        name="9. Hartmann(6)^16_Zero(4)",
        if_permute=True,
        sub_funcs=[[Hartmann6d(6) for _ in range(16)], Zero(4)],
        noisefunc_label_list=[[False for _ in range(16)], True],
    ),
]

TestFuncs500 = [
    AdditiveFunction(
        name="10. Hartmann(6)_Zero(494)",
        if_permute=True,
        sub_funcs=[Hartmann6d(6), Zero(494)],
        noisefunc_label_list=[False, True],
    ),
    AdditiveFunction(
        name="11. RosenBrock(4)^50_Eggholder(2)^50_Zero(200)",
        if_permute=True,
        sub_funcs=[[[Rosenbrock(4), EggHolder(2)] for _ in range(50)], Zero(200)],
        noisefunc_label_list=[[[False, False] for _ in range(50)], True],
    ),
    AdditiveFunction(
        name="12. ProductSines(100)^3_RosenBrock(100)^2",
        if_permute=True,
        sub_funcs=[
            [ProductSines(100) for _ in range(3)],
            [Rosenbrock(100) for _ in range(2)],
        ],
    ),
    AdditiveFunction(
        name="13. Eggholder(2)^250",
        if_permute=True,
        sub_funcs=[EggHolder(2) for _ in range(250)],
    ),
    AdditiveFunction(
        name="14. Rastrigin(500)",
        if_permute=True,
        sub_funcs=[
            Rastrigin(500),
        ],
    ),
    AdditiveFunction(
        name="15. Ackley(10)^50",
        if_permute=True,
        sub_funcs=[Ackley(10) for _ in range(50)],
    ),
    AdditiveFunction(
        name="16. Ackley(20)^5_ProductSines(20)^5_Hartmann6d(6)^25_Noise(150)",
        if_permute=True,
        sub_funcs=[
            [[Ackley(20), ProductSines(20)] for _ in range(5)],
            [Hartmann6d(6) for _ in range(25)],
            Noise(150),
        ],
        noisefunc_label_list=[
            [[False, False] for _ in range(5)],
            [False for _ in range(25)],
            True,
        ],
    ),
    AdditiveFunction(
        name="17. Ackley(20)^5_ProductSines(20)^5_0.5Rastrigin(20)^5_0.2Hartmann6d(6)^30_Noise(20)",
        if_permute=True,
        sub_funcs=[
            [[Ackley(20), ProductSines(20), Rastrigin(20)] for _ in range(5)],
            [Hartmann6d(6) for _ in range(30)],
            Noise(20),
        ],
        weights=[[[1, 1, 0.5] for _ in range(5)], [0.2 for _ in range(30)], 1],
        noisefunc_label_list=[
            [[False, False, False] for _ in range(5)],
            [False for _ in range(30)],
            True,
        ],
    ),
    AdditiveFunction(
        name="18. Ackley(20)^5_ProductSines(20)^5_Rastrigin(20)^5_Hartmann6d(6)^30_Noise(20)",
        if_permute=True,
        sub_funcs=[
            [[Ackley(20), ProductSines(20), Rastrigin(20)] for _ in range(5)],
            [Hartmann6d(6) for _ in range(30)],
            Noise(20),
        ],
        noisefunc_label_list=[
            [[False, False, False] for _ in range(5)],
            [False for _ in range(30)],
            True,
        ],
    ),
    AdditiveFunction(
        name="19. Hartmann(6)^80_Zero(20)",
        if_permute=True,
        sub_funcs=[[Hartmann6d(6) for _ in range(80)], Zero(20)],
        noisefunc_label_list=[[False for _ in range(80)], True],
    ),
]

TestFuncs1000 = [
    AdditiveFunction(
        name="20. Ackley(10)_Zero(990)",
        if_permute=True,
        sub_funcs=[Ackley(10), Zero(990)],
        noisefunc_label_list=[False, True],
    ),
    AdditiveFunction(
        name="21. RosenBrock(4)^100_Eggholder(2)^100_Zero(400)",
        if_permute=True,
        sub_funcs=[[[Rosenbrock(4), EggHolder(2)] for _ in range(100)], Zero(400)],
        noisefunc_label_list=[[[False, False] for _ in range(100)], True],
    ),
    AdditiveFunction(
        name="22. ProductSines(100)^5_RosenBrock(100)^5",
        if_permute=True,
        sub_funcs=[
            [ProductSines(100) for _ in range(5)],
            [Rosenbrock(100) for _ in range(5)],
        ],
    ),
    AdditiveFunction(
        name="23. Eggholder(2)^500",
        if_permute=True,
        sub_funcs=[EggHolder(2) for _ in range(500)],
    ),
    AdditiveFunction(
        name="24. Rastrigin(1000)",
        if_permute=True,
        sub_funcs=[
            Rastrigin(1000),
        ],
    ),
    AdditiveFunction(
        name="25. Ackley(10)^100",
        if_permute=True,
        sub_funcs=[Ackley(10) for _ in range(100)],
    ),
    AdditiveFunction(
        name="26. Ackley(20)^10_ProductSines(20)^10_Hartmann6d(6)^50_Noise(300)",
        if_permute=True,
        sub_funcs=[
            [[Ackley(20), ProductSines(20)] for _ in range(10)],
            [Hartmann6d(6) for _ in range(50)],
            Noise(300),
        ],
        noisefunc_label_list=[
            [[False, False] for _ in range(10)],
            [False for _ in range(50)],
            True,
        ],
    ),
    AdditiveFunction(
        name="27. Ackley(20)^10_ProductSines(20)^10_0.5Rastrigin(20)^10_0.2Hartmann6d(6)^60_Noise(40)",
        if_permute=True,
        sub_funcs=[
            [[Ackley(20), ProductSines(20), Rastrigin(20)] for _ in range(10)],
            [Hartmann6d(6) for _ in range(60)],
            Noise(40),
        ],
        weights=[[[1, 1, 0.5] for _ in range(10)], [0.2 for _ in range(60)], 1],
        noisefunc_label_list=[
            [[False, False, False] for _ in range(10)],
            [False for _ in range(60)],
            True,
        ],
    ),
    AdditiveFunction(
        name="28. Ackley(20)^10_ProductSines(20)^10_Rastrigin(20)^10_Hartmann6d(6)^60_Noise(40)",
        if_permute=True,
        sub_funcs=[
            [[Ackley(20), ProductSines(20), Rastrigin(20)] for _ in range(10)],
            [Hartmann6d(6) for _ in range(60)],
            Noise(40),
        ],
        noisefunc_label_list=[
            [[False, False, False] for _ in range(10)],
            [False for _ in range(60)],
            True,
        ],
    ),
    AdditiveFunction(
        name="29. Hartmann(6)^160_Zero(40)",
        if_permute=True,
        sub_funcs=[[Hartmann6d(6) for _ in range(160)], Zero(40)],
        noisefunc_label_list=[[False for _ in range(160)], True],
    ),
]

TestFuncs = TestFuncs100 + TestFuncs500 + TestFuncs1000

if __name__ == "__main__":
    print(TestFuncs)



from .logging import logger
import numpy as np
import torch
import os
import random


def get_attr(o, n, v):
    attr = getattr(o, n, None)
    if attr is None:
        logger.debug(f"{n} not found, default: {v}")
        attr = v
    return attr


def set_seed(seed=None, torch_deterministic=False):
    if (seed is None) and torch_deterministic:
        seed = 42
    elif seed is None:
        seed = np.random.randint(0, 10000)
    logger.info(f"Setting seed: {seed}")

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    if torch_deterministic:
        # refer to https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.set_deterministic(True)
    else:
        torch.backends.cudnn.benchmark = True
        torch.backends.cudnn.deterministic = False

    return seed


def get_torch_whl(cuda_version="11.8", python_version="3.8", os_version="linux"):
    import requests
    from os.path import join
    from bs4 import BeautifulSoup

    # 构建请求的URL
    request_url = "https://download.pytorch.org/whl/torch/"
    base_url = "https://download.pytorch.org"

    # 发送请求
    response = requests.get(request_url)
    if response.status_code != 200:
        return "Unable to access PyTorch download page, please check your network connection."

    # 解析页面内容
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all("a")

    # 构建筛选条件
    os_filter = "linux" if os_version.lower() == "linux" else "win"
    python_filter = f"cp{python_version.replace('.', '')}"
    cuda_filter = (
        f"cu{cuda_version.replace('.', '')}" if cuda_version != "cpu" else "cpu"
    )

    # 筛选符合条件的链接
    filtered_links = []
    for link in links:
        href = link.get("href")
        if os_filter in href and python_filter in href and cuda_filter in href:
            filtered_links.append(href)

    # 如果没有找到符合条件的链接
    if not filtered_links:
        return "No torch whl file found that meets the specified criteria."

    # 返回最新的链接
    latest_link = sorted(filtered_links, reverse=True)[0]
    return base_url + latest_link


if __name__ == "__main__":
    print(get_torch_whl(cuda_version="11.4", python_version="3.9", os_version="linux"))



import re
import numpy as np
import os
import matplotlib.pyplot as plt
import pandas as pd

from .utils.logging import logger
from HDBOBenchmark.utils.plot import (
    get_linestyle_list,
    cal_std_data,
    plt_init,
    sub_fig,
    plot,
)
from HDBOBenchmark.utils.os import load_pickle

wilcoxon_method = "ALEBO"


def plot_result(func_list, method_list, result_path="./result"):

    max_len = 100

    # init
    plt_init(style="seaborn-paper")  # bmh for common, seaborn-paper for paper
    linestyle_list = get_linestyle_list(len(method_list))

    # cons
    save_dir = "./figs/result/"

    # Params = ({'ymin': -50, 'ymax': 100000},
    #           {},
    #           {'ymin': -200, 'ymax': 1500})
    Params = ({}, {}, {})
    # excel
    excel_data = dict()
    excel_frames = []  # [30, 50, 100]
    for excel_i in excel_frames:
        excel_data[str(excel_i)] = {"func": list()}
        for method in method_list:
            excel_data[str(excel_i)][method] = list()
            excel_data[str(excel_i)][method + "_std"] = list()
    excel_data["wilcoxon"] = {"func": list()}
    for method in method_list:
        excel_data["wilcoxon"][method] = list()

    method_rank_data = {method: [] for method in method_list}
    # loop
    if not isinstance(func_list, list):
        func_list = [func_list]
    for i, func in enumerate(func_list):
        name = func.name
        dim = func.dim
        fig = plt.figure()
        ax = sub_fig(fig, xlabel="Iteration", ylabel="Historical Best Value")
        excel_data["wilcoxon"]["func"].append(name)
        for excel_i in excel_frames:
            excel_data[str(excel_i)]["func"].append(name)

        wilcoxon_data = dict()
        data_max_step = -1
        method_index = 0
        method_data = {method: [] for method in method_list}
        for method in method_list:
            excel_data["wilcoxon"][method].append("N/A")
            for excel_i in excel_frames:
                excel_data[str(excel_i)][method].append("N/A")
                excel_data[str(excel_i)][method + "_std"].append("N/A")
            wilcoxon_data[method] = list()
            load_cnt = 0
            load_path = os.path.join(
                result_path, f"{method}/{name}_{dim}_{load_cnt}.pkl"
            )
            data = list()
            while os.path.exists(load_path):
                result_dict = load_pickle(load_path)[0]
                if isinstance(result_dict, dict):
                    if "best_y_history" in result_dict.keys():
                        best_y_history = result_dict["best_y_history"]
                    elif "y_history" in result_dict.keys():
                        best_y_history = result_dict["y_history"]
                        _y = np.inf
                        for k in range(len(best_y_history)):
                            if best_y_history[k] > _y:
                                best_y_history[k] = _y
                            _y = best_y_history[k]
                            best_y_history[k] = min(best_y_history[k], 100 + 10)
                    else:
                        logger.warning(f"{load_path} doesnt provide a valid dict.")
                    if len(best_y_history) > 100:
                        best_y_history = best_y_history[:max_len]
                    data.append(
                        {
                            "x": np.arange(1, len(best_y_history) + 1),
                            "y": best_y_history,
                            "name": method,
                        }
                    )
                    method_data[method].append(best_y_history)
                    data_max_step = max(data_max_step, len(best_y_history))
                else:
                    logger.warning(f"{load_path} is not in valid format.")
                load_cnt += 1
                load_path = os.path.join(
                    result_path, f"{method}/{name}_{dim}_{load_cnt}.pkl"
                )
            if len(data) == 0:  # no data
                continue
            for d in data:
                wilcoxon_data[method].append(float(d["y"][-1]))
            data = cal_std_data(*data)
            plot(
                data,
                ax,
                color=linestyle_list[method_index][0],
                linestyle=linestyle_list[method_index][1],
            )
            # save excel
            for excel_i in excel_frames:
                excel_data[str(excel_i)][method][-1] = data["y"][excel_i - 1]
                excel_data[str(excel_i)][method + "_std"][-1] = data["std"][excel_i - 1]
            method_index += 1

        if data_max_step == -1:
            continue

        oracle_value = func.min_val
        ax.plot(
            np.arange(data_max_step + 1),
            oracle_value * np.ones(data_max_step + 1),
            f"r-.",
            alpha=0.7,
        )
        if i < len(Params) and "ymin" in Params[i]:
            ax.set_ylim(ymin=Params[i]["ymin"], ymax=Params[i]["ymax"])
            name += "_local"

        # For paper
        ax.set_ylim(ymin=-5, ymax=105)
        # ax.set_yscale("log", base=10)
        title = name
        # name_pre, name_suf = re.match("([0-9]*)\.(.*)", name).groups()
        # title = "No." + name_pre + name_suf
        # if len(title) >= 72:
        #     title = title[:70] + "\n" + title[70:]
        ax.set_title(title, loc="center")

        ax.legend(loc="best", frameon=True, facecolor="white")
        plt.tight_layout()

        # write png to file
        os.makedirs(save_dir, exist_ok=True)
        save_img_path = os.path.join(save_dir, f"{name}.png")
        if os.path.exists(save_img_path):
            os.remove(save_img_path)
        plt.savefig(save_img_path, bbox_inches="tight", dpi=300)
        logger.info(f"Img saved at {save_img_path}.")
        plt.close(fig)

        # Rank calculation
        rank_data = {method: [] for method in method_list}
        for method in method_list:
            if len(method_data[method]) == 0:
                method_data.pop(method)
                method_list.remove(method)
        min_traj_length = min(len(trajs) for trajs in method_data.values())
        for i_traj in range(min_traj_length):
            traj_values = {method: [] for method in method_list}
            for step in range(data_max_step):
                step_values = []
                for method in method_list:
                    if step < len(method_data[method][i_traj]):
                        step_values.append((method, method_data[method][i_traj][step]))
                    else:
                        step_values.append((method, method_data[method][i_traj][-1]))
                step_values.sort(key=lambda x: x[1])
                for rank, (method, _) in enumerate(step_values):
                    traj_values[method].append(rank + 1)
            for method in method_list:
                rank_data[method].append(
                    {
                        "x": np.arange(1, len(traj_values[method]) + 1),
                        "y": traj_values[method],
                        "name": method,
                    }
                )

        plot_rank_data(
            method_list=method_list,
            rank_data=rank_data,
            save_dir=save_dir,
            fig_name=name,
        )
        for method in method_list:
            method_rank_data[method].extend(rank_data[method])

        # write excel
        # for excel_name in excel_data.keys():
        #     df = pd.DataFrame(excel_data[excel_name])
        #     save_excel_path = os.path.join(save_dir, f"data_{excel_name}.csv")
        #     if os.path.exists(save_excel_path):
        #         os.remove(save_excel_path)
        #     df.to_csv(save_excel_path, float_format="%.2f")

    plot_rank_data(
        method_list=method_list,
        rank_data=method_rank_data,
        save_dir=save_dir,
        fig_name="all",
    )


def plot_rank_data(method_list, rank_data, save_dir, fig_name):
    linestyle_list = get_linestyle_list(len(method_list))
    # Plot rank data
    fig = plt.figure()
    ax = sub_fig(fig, xlabel="Iteration", ylabel="Rank")
    method_index = 0
    for method in method_list:
        if len(rank_data[method]) == 0:  # no data
            continue
        data = cal_std_data(*rank_data[method])
        plot(
            data,
            ax,
            color=linestyle_list[method_index][0],
            linestyle=linestyle_list[method_index][1],
        )
        method_index += 1
    ax.legend(loc="best", frameon=True, facecolor="white")
    plt.tight_layout()

    # Save rank plot
    save_rank_path = os.path.join(save_dir, f"{fig_name}_rank.png")
    if os.path.exists(save_rank_path):
        os.remove(save_rank_path)
    plt.savefig(save_rank_path, bbox_inches="tight", dpi=300)
    logger.info(f"Rank img saved at {save_rank_path}.")
    plt.close(fig)


def move_result(cur_dir, target_dir):
    import shutil
    import os

    file_list = os.listdir(cur_dir)
    for f in file_list:
        result = re.match("(.*_)[0-9]*(\.pkl)", f)
        if not result:
            continue
        f_pre, f_suf = result.groups()
        n = 0
        f_new = os.path.join(target_dir, f"{f_pre}{n}{f_suf}")
        while os.path.exists(f_new):
            n += 1
            f_new = os.path.join(target_dir, f"{f_pre}{n}{f_suf}")
        shutil.move(os.path.join(cur_dir, f), f_new)



import torch


def dist2gp(x, c):
    """Calculates squared distance between two sets of points.
    Args:
        x: A 'n_d x ndims' tensor
        c: A 'n_c x ndims' tensor
    Returns:
        A 'n_d x n_c' tensor
    """
    n_d, dim_x = x.shape
    n_c, dim_c = c.shape
    if dim_x != dim_c:
        raise Exception("Data dimension does not match dimension of centers.")
    xx = torch.sum(x**2, axis=1)  # (n_d,)
    cc = torch.sum(c**2, axis=1)  # (n_c,)
    xc = x @ c.T  # (n_data, n_centres)
    return (
        torch.repeat_interleave(xx.unsqueeze(-1), n_c, dim=1)
        + torch.repeat_interleave(cc.unsqueeze(0), n_d, dim=0)
        - 2.0 * xc
    )


if __name__ == "__main__":
    from HDBOBenchmark.utils.logging import logger

    x = torch.tensor([[1, 2, 3, 4, 5], [11, 12, 13, 14, 15]])
    logger.info(dist2gp(x, x))



from .FunctionBase import BOTorchTestFunction
from botorch.test_functions import EggHolder as EggHolder_


class EggHolder(BOTorchTestFunction):
    def __init__(
        self,
        dim,
        x_scaler=None,
        y_scaler=None,
        auto_init_x_scaler=True,
        auto_init_y_scaler=True,
    ):
        super().__init__(EggHolder_(), "EggHolder")
        self.y_scale = 2000
        self.init_scaler(x_scaler, y_scaler, auto_init_x_scaler, auto_init_y_scaler)



def plot_func(Func):
    import os
    import numpy as np
    import matplotlib.pyplot as plt

    # create function instance (dim=2)
    func = Func(dim=2)
    assert func.dim == 2

    # create meshgrids
    meshgrid_fineness = 100
    bounds = func.bound
    x = np.linspace(bounds[0][0], bounds[0][1], meshgrid_fineness)
    y = np.linspace(bounds[1][0], bounds[1][1], meshgrid_fineness)
    X, Y = np.meshgrid(x, y)

    # calculate the function value of each point on the grid
    Z = np.array(
        [func.evaluate(np.array([x, y]))["y"] for x, y in zip(np.ravel(X), np.ravel(Y))]
    )
    Z = Z.reshape(X.shape)

    # using Matplotlib to draw a surface graph, allow dragging for a 3D-view
    fig = plt.figure()
    ax = fig.gca(projection="3d")  # ax = fig.add_axes(Axes3D(fig))
    # ax.contour(X, Y, Z, cmap='jet', levels=20)
    ax.plot_surface(X, Y, Z, cmap=plt.cm.jet)
    ax.set_xlabel("X-axis")
    ax.set_ylabel("Y-axis")
    ax.set_zlabel("Z-axis")
    plt.title(f"{Func.__name__} Function")
    plt.show()

    # save to local
    os.makedirs("./figs/", exist_ok=True)
    fig.savefig(f"./figs/{Func.__name__}.png")


if __name__ == "__main__":
    from HDBOBenchmark import Sphere as Func

    plot_func(Func)



import torch
from torch import Tensor


def columnwise_clamp(
    X: Tensor, lower: Tensor, upper: Tensor, raise_on_violation=False
) -> Tensor:
    """Clamp values of a ndarray in column-wise fashion (with support for t-batches).

    Args:
        X: The 'b x n x d' input tensor. If 2-dimensional, 'b' is assumed to be 1.
        lower: The column-wise lower bounds. If scalar, apply bound to all columns.
        upper: The column-wise upper bounds. If scalar, apply bound to all columns.
        raise_on_violation: If `True`, raise an exception when the elments in `X`
            are out of the specified bounds (up to numerical accuracy). This is
            useful for post-processing candidates generated by optimizers that
            satisfy imposed bounds only up to numerical accuracy.

    Returns:
        The clamped ndarray.
    """
    min_bounds = lower.expand_as(X)  # np.broadcast_to(lower, X.shape)
    max_bounds = upper.expand_as(X)  # np.broadcast_to(upper, X.shape)
    if torch.any(min_bounds > max_bounds):
        raise ValueError("Minimum values must be <= maximum values")
    Xout = X.max(min_bounds)  # np.maximum(X, min_bounds)
    Xout = Xout.min(max_bounds)  # np.minimum(Xout, max_bounds)
    if raise_on_violation and not torch.allclose(Xout, X):
        raise RuntimeError("Original value(s) are out of bounds.")
    return Xout
